###########################################################################
#
# docsave.pm
# A component of the Greenstone digital library software
# from the New Zealand Digital Library Project at the 
# University of Waikato, New Zealand.
#
# Copyright (C) 1999 New Zealand Digital Library Project
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
#
###########################################################################

# This document processor saves a document in the 
# archives directory of a collection (as xml)

use strict;
no strict 'refs';

package docsave;

eval {require bytes};

use arcinfo;
use expinfo;
use docproc;
use util;


sub BEGIN {
    @docsave::ISA = ('docproc');
}

sub new {
    my ($class, $collection, $info, $verbosity, 
	$gzip, $groupsize, $outhandle, $service, $saveas) = @_;
    my $self = new docproc ();
    
    $groupsize=1 unless defined $groupsize;
    $service="import" unless defined $service;

    $self->{'collection'} = $collection;
    if ($service eq "import"){
	$self->{'archive_info'} = $info;
    } elsif ($service eq "export"){ 
	$self->{'export_info'} = $info;
    } else {
	return;
    }

    $self->{'verbosity'} = $verbosity;
    $self->{'gzip'} = $gzip;
    $self->{'keepimportstructure'} = 0;
    $self->{'groupsize'} = $groupsize;
    $self->{'gs_count'} = 0;

    $self->{'outhandle'} = 'STDERR';
    $self->{'outhandle'} = $outhandle if defined $outhandle;
    $self->{'service'} = $service;
    $self->{'saveas'} = $saveas;

    # set a default for the archive directory
    if ($service eq "import"){
	$self->{'archive_dir'} = &util::filename_cat ($ENV{'GSDLCOLLECTDIR'}, "archives");
    } elsif ($service eq "export") {
	# set a default for the export directory
	$self->{'export_dir'} = &util::filename_cat($ENV{'GSDLCOLLECTDIR'}, "export");
    } else {
	return;
    }
    $self->{'sortmeta'} = undef;
    
    return bless $self, $class;
}

sub setarchivedir {
    my $self = shift (@_);
    my ($archive_dir) = @_;
    
    &util::mk_all_dir ($archive_dir) unless -e $archive_dir;
    $self->{'archive_dir'} = $archive_dir;
}

sub setexportdir {
    my $self = shift (@_);
    my ($export_dir) = @_;
 
    &util::mk_all_dir ($export_dir) unless -e $export_dir;
    $self->{'export_dir'} = $export_dir;
}

sub set_sortmeta {
    my $self = shift (@_);
    my ($sortmeta, $removeprefix, $removesuffix) = @_;
    
    $self->{'sortmeta'} = $sortmeta;
    if (defined ($removeprefix) && $removeprefix ) {
	$removeprefix =~ s/^\^//; # don't need a leading ^
	$self->{'removeprefix'} = $removeprefix;
    }
    if (defined ($removesuffix) && $removesuffix) {
	$removesuffix =~ s/\$$//; # don't need a trailing $
	$self->{'removesuffix'} = $removesuffix;
    }
}

sub process {
    my $self = shift (@_);
    my ($doc_obj) = @_;
 
    my $outhandle = $self->{'outhandle'};
    my $service = $self->{'service'} || "import";

    # Define the SaveAs Type
    my $save_as = $self->{'saveas'} || "GA";
    my $collection = $self->{'collection'};

    if ($self->{'groupsize'} > 1) {
	$self->group_process ($doc_obj);
	return;
    }

    my $OID = $doc_obj->get_OID();
    $OID = "NULL" unless defined $OID;

    # get document's directory
    my $doc_dir = $self->get_doc_dir ($OID, $doc_obj->get_source_filename());

    # groupsize is 1 (i.e. one document per XML file) so sortmeta 
    # may be used

    if ($service eq "import") {
	my $archive_info = $self->{'archive_info'};
    } elsif ($service eq "export") {
	my $export_info = $self->{'export_info'};
    } else {
	return;
    }
    
    # copy all the associated files, add this information as metadata
    # to the document
    if ($service eq "export" && $save_as eq "DSpace") {
	# open contents file
	my $doc_contents_file
	    = &util::filename_cat ($self->{'export_dir'},$doc_dir, "contents");
	
	if (!open(OUTDOC_EXPORT_CONTENTS,">$doc_contents_file")){
	    print $outhandle "docsave::process could not write collection contents to file $doc_contents_file\n";
	    return;
	} 
	$self->process_assoc_files ($doc_obj, $doc_dir, 'docsave::OUTDOC_EXPORT_CONTENTS');
    } else {
	$self->process_assoc_files ($doc_obj, $doc_dir, '');
    }
        
    my $doc_file;
    my $doc_mets_file;
    my $doc_txt_file;
    my $short_doc_file;

    #Import collection to GS2 in GS Archive format and METs format
    if ($service eq "import") {
	my $doc_file 
	    = &util::filename_cat ($self->{'archive_dir'}, $doc_dir, "doc.xml");

	#***define doctxt.xml file
	my $doc_txt_file
	    = &util::filename_cat ($self->{'archive_dir'}, $doc_dir,"doctxt.xml");

	my $import_working_dir
	    =&util::filename_cat ($self->{'archive_dir'}, $doc_dir);
	
	#***define docmets.xml file
	my $doc_mets_file
	    = &util::filename_cat ($self->{'archive_dir'},$doc_dir, "docmets.xml");

	if ($save_as eq "GA") {
	    $short_doc_file = util::filename_cat ($doc_dir, "doc.xml");
	} elsif ($save_as eq "METS") {
	    #my $short_txt_doc_file=&util::filename_cat ($doc_dir, "doctxt.xml");
	    $short_doc_file = &util::filename_cat ($doc_dir, "docmets.xml");
	} else {
	    return;
	}
	
	if ($save_as eq "GA") {
	    if (!open (OUTDOC, ">$doc_file")) {
		print $outhandle "docsave::process could not write to file $doc_file\n";
		return;
	    }
	    # save this document
	    $self->output_xml_header('docsave::OUTDOC');
	    $doc_obj->output_section('docsave::OUTDOC',
				 $doc_obj->get_top_section());
	    $self->output_xml_footer('docsave::OUTDOC');
	    
	    close OUTDOC;
	} elsif ($save_as eq "METS") {
	    # save the document without metadata:doctxt.xml
	    
	    if (!open(OUTDOC_TXT, ">$doc_txt_file")){
		print $outhandle "docsave::process could not write to file $doc_txt_file\n";
		return;
	    }
	    
	    $self->output_txt_xml_header('docsave::OUTDOC_TXT');
	    $doc_obj->output_txt_section('docsave::OUTDOC_TXT', $doc_obj->get_top_section());
	    #$self->output_txt_xml_footer('docsave::OUTDOC_TXT');
	    
	    # Convert doctxt.xml file to docmets.xml
	    if (!open(OUTDOC_METS,">$doc_mets_file")){
		print $outhandle "docsave::process could not write to file $doc_mets_file\n";
		return;
	    }
	    
	    $self->output_mets_xml_header('docsave::OUTDOC_METS', $OID);
	    $doc_obj->output_mets_section('docsave::OUTDOC_METS',
					  $doc_obj->get_top_section());
	    $self->output_mets_xml_footer('docsave::OUTDOC_METS');
	    
	    close OUTDOC_TXT;
	    close OUTDOC_METS;
	} else { # save_as isn't GA or METS
	    print $outhandle "docsave::process unrecognised saveas type, $save_as\n";
	    return;
	}
    }
    
    ## Export the collection to METs format or DSpace Archive Format into the export directory
    if ($service eq "export") {
	my $doc_dc_file;
	my $doc_contents_file;

	my $export_working_dir
	    =&util::filename_cat ($self->{'export_dir'}, $doc_dir);
       
	if ($save_as eq "METS") {
	    $doc_mets_file
		= &util::filename_cat ($self->{'export_dir'},$doc_dir, "docmets.xml");
	    
	    $doc_txt_file
		= &util::filename_cat ($self->{'export_dir'},$doc_dir, "doctxt.xml");
	    
	    if (!open(OUTDOC_EXPORT_TXT, ">$doc_txt_file")){
		print $outhandle "docsave::process could not write TXT to file $doc_txt_file\n";
		return;
	    }
	    
	    $self->output_txt_xml_header('docsave::OUTDOC_EXPORT_TXT');
	    $doc_obj->output_txt_section('docsave::OUTDOC_EXPORT_TXT', $doc_obj->get_top_section());
	    
	    if (!open(OUTDOC_EXPORT_METS,">$doc_mets_file")){
		print $outhandle "docsave::process could not write METS format to file $doc_mets_file\n";
		return;
	    }
	    $self->output_mets_xml_header('docsave::OUTDOC_EXPORT_METS', $OID);
	    $doc_obj->output_mets_section('docsave::OUTDOC_EXPORT_METS',$doc_obj->get_top_section(), $export_working_dir);
	    $self->output_mets_xml_footer('docsave::OUTDOC_EXPORT_METS');
    
	    close OUTDOC_EXPORT_TXT;
	    close OUTDOC_EXPORT_METS;
	} elsif ($save_as eq "DSpace") {

	    # Generate dublin_core.xml file
	    $doc_dc_file
		= &util::filename_cat ($self->{'export_dir'},$doc_dir, "dublin_core.xml");
   
	    if (!open(OUTDOC_EXPORT_DC,">$doc_dc_file")){
		print $outhandle "docsave::process could not write dublin core to file $doc_dc_file\n";
		return;
	    }   
	   
	    $self->output_dc_xml_header('docsave::OUTDOC_EXPORT_DC', $OID);
	    $doc_obj->output_dc_section('docsave::OUTDOC_EXPORT_DC',$doc_obj->get_top_section(), $export_working_dir);
	    $self->output_dc_xml_footer('docsave::OUTDOC_EXPORT_DC');
    
	    close OUTDOC_EXPORT_DC;
	    close OUTDOC_EXPORT_CONTENTS;
	} else { # save_as isn't METS or DSpace
	    print $outhandle "docsave::process unrecognised saveas type, $save_as\n";
	    return;
	}

	if ($save_as eq "METS") {
	    $short_doc_file = util::filename_cat ($doc_dir, "docmets.xml");
	} elsif ($save_as eq "DSpace") {
	    #my $short_txt_doc_file=&util::filename_cat ($doc_dir, "doctxt.xml");
	    $short_doc_file=&util::filename_cat ($doc_dir, "dublin_core.xml");
	} else {
	    return;
	}
	
    }
    #save for later (for close_file_output())
    $self->{'short_doc_file'} = $short_doc_file;	
 
    if ($self->{'gzip'}) {
	my $doc_file = $self->{'gs_filename'};
	`gzip $doc_file`;
	$doc_file .= ".gz";
	$short_doc_file .= ".gz";
	if (!-e $doc_file) {
	    print $outhandle "error while gzipping: $doc_file doesn't exist\n";
	    return 0;
	}
    }

    # do the sortmeta thing
    my ($metadata);	
    if (defined ($self->{'sortmeta'})) {
	$metadata = $doc_obj->get_metadata_element($doc_obj->get_top_section(),
						   $self->{'sortmeta'});
    }
    if (defined ($metadata) && $metadata) {
	# do remove prefix/suffix
	if (defined($self->{'removeprefix'})) {
	    $metadata =~ s/^$self->{'removeprefix'}//;
	}
	if (defined($self->{'removesuffix'})) {
	    $metadata =~ s/$self->{'removesuffix'}$//;
	}
	$metadata = &sorttools::format_metadata_for_sorting($self->{'sortmeta'}, $metadata, $doc_obj);
    }
    # store reference in the archive_info and export_info
    if ($service eq "export") {
	$self->{'export_info'}->add_info($OID, $short_doc_file, $metadata);	
    } elsif ($service eq "import") {
	$self->{'archive_info'}->add_info($OID, $short_doc_file, $metadata);
    }
}

      
sub group_process {
    my $self = shift (@_);
    my ($doc_obj) = @_;
    
    my $outhandle = $self->{'outhandle'};

    my $OID = $doc_obj->get_OID();
    $OID = "NULL" unless defined $OID;

    my $groupsize = $self->{'groupsize'};
    my $gs_count = $self->{'gs_count'};
    my $open_new_file = (($gs_count % $groupsize)==0);

    # opening a new file, or document has assoicated files => directory needed
    if (($open_new_file) || (scalar(@{$doc_obj->get_assoc_files()})>0)) {

	# get document's directory
	my $doc_dir = $self->get_doc_dir ($OID, $doc_obj->get_source_filename());

	# copy all the associated files, add this information as metadata
	# to the document
	$self->process_assoc_files ($doc_obj, $doc_dir);


	if ($open_new_file) {
	    # only if opening new file
	    my $doc_file 
		= &util::filename_cat ($self->{'archive_dir'}, $doc_dir, "doc.xml");
	    my $short_doc_file = &util::filename_cat ($doc_dir, "doc.xml");
	    
	    if ($gs_count>0)
	    {
		return if (!$self->close_file_output());
	    }
	    
	    if (!open (OUTDOC, ">$doc_file")) {
		print $outhandle "docsave::group_process could not write to file $doc_file\n";
		return;
	    }
	    $self->{'gs_filename'} = $doc_file;
	    $self->{'gs_short_filename'} = $short_doc_file;
	    $self->{'gs_OID'} = $OID;

	    $self->output_xml_header('docsave::OUTDOC');
	}
    }

    # save this document
    $doc_obj->output_section('docsave::OUTDOC', $doc_obj->get_top_section());

    $self->{'gs_count'}++;
}

sub get_doc_dir {
    my $self = shift (@_);
    my ($OID, $source_filename) = @_;
    my $doc_info;
    my $doc_dir = '';
    my $service = $self-> {'service'};
    my $working_dir;
    my $working_info;

    if ($service eq "import") {
	$doc_info = $self->{'archive_info'}->get_info($OID);
	$working_dir = $self->{'archive_dir'};
	$working_info = $self->{'archive_info'};
    } elsif ($service eq "export") {
	$doc_info =$self->{'export_info'}->get_info($OID);
        $working_dir = $self->{'export_dir'};
	$working_info = $self->{'export_info'};
    } else {
	return;
    }
    if (defined $doc_info && scalar(@$doc_info) >= 1) {
	# this OID already has an assigned directory, use the
	# same one.
	$doc_dir = $doc_info->[0];
	$doc_dir =~ s/\/?doc(mets)?\.xml(\.gz)?$//;
    } elsif ($self->{'keepimportstructure'}) {
	$source_filename = &File::Basename::dirname($source_filename);
	$source_filename =~ s/[\\\/]+/\//g;
	$source_filename =~ s/\/$//;
 
	
	#print STDERR "Source filename: $source_filename; \nImport dir:",$ENV{'GSDLIMPORTDIR'}, "\n";
	$doc_dir = substr($source_filename, length($ENV{'GSDLIMPORTDIR'}) + 1);

    }
    if ($doc_dir eq "") {
	# have to get a new document directory
	my $doc_dir_rest = $OID;
	my $doc_dir_num = 0;
	do {
	    $doc_dir .= "/" if $doc_dir_num > 0;
	    if ($doc_dir_rest =~ s/^(.{1,8})//) {
		$doc_dir .= $1;
		$doc_dir_num++;
	    }
	} while ($doc_dir_rest ne "" && 
		 ((-d &util::filename_cat ($working_dir, "$doc_dir.dir")) || 
		  ($working_info->size() >= 1024 && $doc_dir_num < 2)));

	$doc_dir .= ".dir";
	&util::mk_all_dir (&util::filename_cat ($working_dir, $doc_dir));
    }
    return $doc_dir;
}

sub process_assoc_files {
    my $self = shift (@_);
    my ($doc_obj, $doc_dir, $handle) = @_;

    my $outhandle = $self->{'outhandle'};
    
    my @assoc_files = ();
    my $filename;;
    my $working_dir;
    my $service = $self->{'service'};
    my $save_as = $self->{'saveas'};

    if ($service eq "import") {
	$working_dir = $self->{'archive_dir'};
    } elsif ($service eq "export"){
	$working_dir = $self->{'export_dir'};
    } else {
	return;
    }

    if ($save_as eq "DSpace") {

	my ($tail_filename) = ($doc_obj->get_source_filename() =~ m/\/([^\/\\]*)$/);

	print $handle "$tail_filename\n";

	$filename = &util::filename_cat($working_dir, $doc_dir, $tail_filename);
	&util::hard_link ($doc_obj->get_source_filename(), $filename);
    }
           
    foreach my $assoc_file (@{$doc_obj->get_assoc_files()}) {
	my ($dir, $afile) = $assoc_file->[1] =~ /^(.*?)([^\/\\]+)$/;
	$dir = "" unless defined $dir;

	# Store the associated file to the "contents" file
	if ($save_as eq "DSpace") {
	    print $handle "$assoc_file->[1]\n";
	}

	if (-e $assoc_file->[0]) {
	    $filename = &util::filename_cat($working_dir, $doc_dir, $afile);

	    &util::hard_link ($assoc_file->[0], $filename);
	    
	    $doc_obj->add_utf8_metadata ($doc_obj->get_top_section(),
					 "gsdlassocfile",
					 "$afile:$assoc_file->[2]:$dir");
	    $doc_obj->set_utf8_metadata_element ($doc_obj->get_top_section(),
						 "assocfilepath",
						 "$doc_dir");
	} elsif ($self->{'verbosity'} > 2) {
	    print $outhandle "docsave::process couldn't copy the associated file " .
		"$assoc_file->[0] to $afile\n";
	}
    }
}
    

sub close_file_output
{
    my ($self) = @_;
    my $service =$self->{'service'};

    # make sure that the handle has been opened - it won't be if we failed
    # to import any documents...
    if (defined(fileno(docsave::OUTDOC))) {
	$self->output_xml_footer('docsave::OUTDOC');    
	close OUTDOC;
    }

    my $OID = $self->{'gs_OID'};
    my $short_doc_file;
    # can we use 'short_doc_file' for GA too?
    if (exists($self->{'saveas'}) && $self->{'saveas'} eq "METS") {
	$short_doc_file=$self->{'short_doc_file'};
    } elsif ($self->{'saveas'} eq "GA") { # "GA"
	$short_doc_file=$self->{'gs_short_filename'};
    } else { # "DSpace"
    }
    
    if ($self->{'gzip'}) {
	my $doc_file = $self->{'gs_filename'};
	`gzip $doc_file`;
	$doc_file .= ".gz";
	$short_doc_file .= ".gz";
	if (!-e $doc_file) {
	    my $outhandle = $self->{'outhandle'};
	    print $outhandle "error while gzipping: $doc_file doesn't exist\n";
	    return 0;
	}
    }

    # store reference in the archive_info and export_infor
    if ($service eq "import") {
	$self->{'archive_info'}->add_info($OID, $short_doc_file);
    } elsif ($service eq "export") {
	$self->{'export_info'}->add_info($OID, $short_doc_file);
    } else {
	return;
    }
    return 1;
}

sub output_xml_header {
    my $self = shift (@_);
    my ($handle) = @_;

    print $handle '<?xml version="1.0" encoding="UTF-8" standalone="no"?>' . "\n";
   
    print $handle '<!DOCTYPE Archive SYSTEM "http://greenstone.org/dtd/Archive/1.0/Archive.dtd">' . "\n";
    print $handle "<Archive>\n";
}

sub output_xml_footer {
    my $self = shift (@_);
    my ($handle) = @_;

    print $handle "</Archive>\n";
}

sub output_txt_xml_header{
    my $self = shift (@_);
    my ($handle) = @_;
    print $handle '<?xml version="1.0" encoding="UTF-8" standalone="no"?>' . "\n";
    print $handle '<!DOCTYPE Archive SYSTEM "http://greenstone.org/dtd/Archive/1.0/Archive.dtd">' . "\n";
}

sub output_txt_xml_footer{
    my $self = shift(@_);
    my ($handle) = @_;
    print $handle "<the end of the file>\n";
}

sub output_mets_xml_header(){
    my $self = shift(@_);
    my ($handle, $OID) = @_;

    print $handle '<?xml version="1.0" encoding="UTF-8" standalone="no"?>' . "\n";
    print $handle '<mets:mets xmlns:mets="http://www.loc.gov/METS/"' . "\n";
    print $handle '           xmlns:xsi="http://www.w3.org/2001/XMLSchema-instance"' . "\n";
    print $handle '           xmlns:gsdl3="http://www.greenstone.org/namespace/gsdlmetadata/1.0/"' . "\n";
    print $handle '           xmlns:xlink="http://www.w3.org/TR/xlink"' ."\n";
    print $handle '           xsi:schemaLocation="http://www.loc.gov/METS/' . "\n";
    print $handle '           http://www.loc.gov/standards/mets/mets.xsd' . "\n";
    print $handle '           http://www.greenstone.org/namespace/gsdlmetadata/1.0/' . "\n";
    print $handle '           http://www.greenstone.org/namespace/gsdlmetadata/1.0/gsdl_metadata.xsd"' . "\n";
    print $handle '           OBJID="'. $OID. ':2">' . "\n";
}

sub output_mets_xml_footer() {
    my $self = shift(@_);
    my ($handle) = @_;
    print $handle '</mets:mets>' . "\n";
}

sub output_dc_xml_header(){
    my $self = shift(@_);
    my ($handle, $OID) = @_;

    print $handle '<?xml version="1.0" encoding="UTF-8" standalone="no"?>' . "\n";
#    print $handle '<!DOCTYPE Archive SYSTEM "http://greenstone.org/dtd/Archive/1.0/Archive.dtd">'."\n";
    print $handle '<dublin_core>' . "\n";
}

sub output_dc_xml_footer() {
    my $self = shift(@_);
    my ($handle) = @_;
    print $handle '</dublin_core>' . "\n";
}
1;
